package com.outr.webcommunicator

import java.util.UUID
import org.powerscala.Logging
import org.powerscala.concurrent.Time._
import java.util.concurrent.locks.ReentrantLock
import org.powerscala.event.Listenable

import org.powerscala.json._

/**
 * Communicator represents how the server communicates to the client and receives messages from the client.
 *
 * Implementations need only provide a receive(message: String) method.
 *
 * @author Matt Hicks <mhicks@powerscala.org>
 */
trait Communicator extends Logging with Listenable {
  private var _lastUpdated = System.currentTimeMillis()

  private val outgoingLock = new ReentrantLock()
  private val incomingLock = new ReentrantLock()

  private var _disposed = false
  private var outgoingCounter = 0
  private var incomingCount = 0

  private var outgoingQueue = Vector.empty[IdentifiableMessage]

  protected[webcommunicator] var communicatorReceiver: CommunicatorReceiver[_] = null
  protected[webcommunicator] var communicatorSender: CommunicatorSender[_] = null

  /**
   * Universally unique identifier for this Communicator instance. Only one Communicator currently loaded can have this
   * id.
   */
  def id: UUID

  /**
   * Called when a message is inbound from the client.
   */
  def receive(event: String, message: String): Unit

  /**
   * Called when a Communicator times out.
   */
  def dispose(): Unit = {
    _disposed = true
  }

  /**
   * True if this communicator has been disposed.
   */
  def disposed = _disposed

  /**
   * How long to wait before disposing the Communicator after the last update is received.
   *
   * Defaults to 30 seconds.
   */
  def timeout: Double = 30.seconds

  /**
   * The amount of time in seconds since the last update.
   */
  final def lastUpdated = fromMillis(System.currentTimeMillis() - _lastUpdated)

  protected[webcommunicator] def updated() = _lastUpdated = System.currentTimeMillis()
  protected[webcommunicator] def outgoing(lastOutgoingId: Int): Seq[IdentifiableMessage] = outgoingSynchronized {
    if (outgoingQueue.nonEmpty) {
      val head = outgoingQueue.head
      if (head.id <= lastOutgoingId) {
        // Client verified receipt of this id - remove and continue
        outgoingQueue = outgoingQueue.tail
        outgoing(lastOutgoingId)
      } else {
        outgoingQueue
      }
    } else {
      Nil
    }
  }
  protected[webcommunicator] def incoming(msg: IdentifiableMessage) = incomingSynchronized {
    val expected = incomingCount + 1
    if (msg.id < expected) {
      error("%s: Invalid incoming id! Expected: %s, Received: %s, Message: %s".format(id, expected, msg.id, msg.message))
    } else {
      if (msg.id > expected) {
        warn("%s: Skipped an incoming id! Expected: %s, Received: %s, Message: %s".format(id, expected, msg.id, msg.message))
      }
      incomingCount = msg.id
      receive(msg.event, msg.message)
    }
  }

  def update() = {
    try {
      if (communicatorReceiver != null) {
        communicatorReceiver.communicatorUpdating()
      }
      if (communicatorSender != null) {
        communicatorSender.communicatorUpdating()
      }
    } catch {
      case t: Throwable => t.printStackTrace()
    }
  }

  /**
   * Asynchronously sends a message to the client.
   */
  def send(event: String, message: String) = outgoingSynchronized {
    outgoingCounter += 1
    outgoingQueue = outgoingQueue :+ IdentifiableMessage(outgoingCounter, event, message)
    if (communicatorReceiver != null) {
      communicatorReceiver.outgoingEnqueued()
    }
    if (communicatorSender != null) {
      communicatorSender.outgoingEnqueued()
    }
  }

  /**
   * Synchronizes against all outgoing actions to execute the supplied function.
   */
  def outgoingSynchronized[T](f: => T) = {
    outgoingLock.lock()
    try {
      f
    } finally {
      outgoingLock.unlock()
    }
  }

  /**
   * Synchronizes against all incoming actions to execute the supplied function.
   */
  def incomingSynchronized[T](f: => T) = {
    incomingLock.lock()
    try {
      f
    } finally {
      incomingLock.unlock()
    }
  }
}

trait CommunicatorListener[C <: Communicator] {
  def manager: CommunicatorManager[C]
  var communicator: C = _

  def communicatorChanged() = {}

  def outgoingEnqueued() = {}

  def communicatorUpdating() = {}
}

trait CommunicatorReceiver[C <: Communicator] extends CommunicatorListener[C] {
  var lastSentId = 0

  protected final def handle(event: Event) = event.event match {
    case "create" => {
      val id = UUID.fromString(event.message)
      communicator = manager.create(id)
      communicator.communicatorReceiver = this
      communicatorChanged()
    }
    case "reconnect" => {
      val reconnect = parse[Reconnect](event.message)
      val id = UUID.fromString(reconnect.uuid)
      lastSentId = reconnect.lastReceiveId
      manager(id) match {
        case Some(c) => {
          communicator = c
          communicator.communicatorReceiver = this
          communicatorChanged()
        }
        case None => println("*** CommunicatorReceiver.handle - Unsupported reconnect when Communicator not found by id")
      }
    }
    case _ if (communicator != null) => communicator.receive(event.event, event.message)
    case _ => println("Attempting to handle: %s when no communicator is defined! (%s)".format(event, this))
  }
}

trait CommunicatorSender[C <: Communicator] extends CommunicatorListener[C] {
}

case class Event(event: String, message: String)

case class Reconnect(uuid: String, lastReceiveId: Int)

case class SendEvent(uuid: String, messages: List[IdentifiableMessage])