package com.outr.webcommunicator.netty

import handler.RequestHandler
import org.jboss.netty.bootstrap.ServerBootstrap
import org.jboss.netty.channel.socket.nio.NioServerSocketChannelFactory
import java.util.concurrent.Executors
import java.net.{URL => JURL, InetSocketAddress}
import org.jboss.netty.channel._
import org.jboss.netty.handler.codec.http._
import org.jboss.netty.handler.stream.ChunkedWriteHandler
import scala.Some
import org.jboss.netty.buffer.ChannelBuffers
import org.jboss.netty.util.CharsetUtil
import annotation.tailrec
import java.util.concurrent.atomic.AtomicBoolean
import org.powerscala.Logging
import com.outr.webcommunicator.{URL, Method}
import java.nio.channels.ClosedChannelException

/**
 * @author Matt Hicks <mhicks@powerscala.org>
 */
trait NettyWebapp extends Webapp with ChannelPipelineFactory with Logging {
  NettyWebapp.instance = this

  private var channels = Map.empty[String, Channel]
  private var resources = List.empty[WebResource]

  val bootstrap = new ServerBootstrap(new NioServerSocketChannelFactory(Executors.newCachedThreadPool(), Executors.newCachedThreadPool()))
  bootstrap.setPipelineFactory(this)

  private val initialized = new AtomicBoolean(false)

  /**
   * Creates a URL from the given values. This method allows overriding the internally seen URL.
   */
  def createURL(method: Method,
                protocol: String,
                host: String,
                port: Int,
                path: String,
                parameters: Map[String, List[String]]) = {
    URL(method, protocol, host, port, path, parameters)
  }

  def bind(host: String, port: Int) = synchronized {
    val address = host match {
      case null => new InetSocketAddress(port)
      case _ => new InetSocketAddress(host, port)
    }
    channels += "%s:%s".format(host, port) -> bootstrap.bind(address)
    if (initialized.compareAndSet(false, true)) {
      initialize()
    }
  }

  def unbind(host: String, port: Int) = synchronized {
    val key = "%s:%s".format(host, port)
    channels.get(key) match {
      case Some(channel) => {
        channel.unbind()
        channels -= key
      }
      case None =>
    }
  }

  def shutdown() = synchronized {
    channels.foreach {
      case (key, channel) => {
        channel.unbind()
        channels -= key
      }
    }
  }

  def register(resource: WebResource): Unit = if (!resources.contains(resource)) synchronized {
    resources = (resource :: resources).sortBy(h => h.priority.value)
  }

  def register(path: String, resource: String): Unit = register(URLResource(path, resource))

  def register(path: String, url: JURL): Unit = register(URLResource(path, url))

  def register(resource: String): Unit = register(URLResource(if (resource.startsWith("/")) {
    resource
  } else {
    "/%s".format(resource)
  }, resource))

  def getPipeline = {
    val pipeline = Channels.pipeline()

    pipeline.addLast("decoder", new HttpRequestDecoder)
    pipeline.addLast("aggregator", new HttpChunkAggregator(Int.MaxValue))
    pipeline.addLast("encoder", new HttpResponseEncoder)
    pipeline.addLast("deflater", new HttpContentCompressor)
    pipeline.addLast("streamer", new ChunkedWriteHandler)
    pipeline.addLast("handler", new ConnectionHandler)

    pipeline
  }

  protected def notFound(context: ChannelHandlerContext, event: MessageEvent) = {
    val response = RequestHandler.createResponse(status = HttpResponseStatus.NOT_FOUND, contentType = "text/plain; charset=UTF-8")
    response.setContent(ChannelBuffers.copiedBuffer(response.getStatus.toString, CharsetUtil.UTF_8))
    context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  protected def forbidden(context: ChannelHandlerContext, event: MessageEvent) = {
    val response = RequestHandler.createResponse(status = HttpResponseStatus.FORBIDDEN, contentType = "text/plain; charset=UTF-8")
    response.setContent(ChannelBuffers.copiedBuffer(response.getStatus.toString, CharsetUtil.UTF_8))
    context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  protected def error(context: ChannelHandlerContext, event: MessageEvent, t: Throwable) = {
    t.printStackTrace()
    val response = RequestHandler.createResponse(status = HttpResponseStatus.INTERNAL_SERVER_ERROR, contentType = "text/plain; charset=UTF-8")
    response.setContent(ChannelBuffers.copiedBuffer(response.getStatus.toString, CharsetUtil.UTF_8))
    // TODO: output additional information
    context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  @tailrec
  private def lookupHandler(context: ChannelHandlerContext, event: MessageEvent, resources: List[WebResource]): Option[RequestHandler] = {
    if (resources.isEmpty) {
      None
    } else {
      val resource = resources.head
      resource.request(this, context, event) match {
        case Some(handler) => Some(handler)
        case None => lookupHandler(context, event, resources.tail)
      }
    }
  }

  protected def contextualize[T](context: ChannelHandlerContext, event: MessageEvent)(f: => T): T = {
    f
  }

  protected def invokeHandler(context: ChannelHandlerContext, event: MessageEvent, handler: RequestHandler) = {
    handler(this, context, event)
  }

  protected def exceptionCaught(context: ChannelHandlerContext, t: Throwable, handler: RequestHandler) = {
    if (t.getMessage == "An existing connection was forcibly closed by the remote host") {
      warn(t.getMessage)    // This happens when the browser is closed, so we can generally ignore it
    } else if (t.getMessage == "An established connection was aborted by the software in your host machine") {
      warn(t.getMessage)     // This happens when the connection is terminated abruptly on page change
    } else if (t.isInstanceOf[ClosedChannelException]) {
      trace("Channel closed abruptly")
    } else if (handler == null) {
      new RuntimeException("ConnectionHandler.exceptionCaught without handler", t).printStackTrace()
    } else {
      handler.errorThrown(t)
    }
  }

  class ConnectionHandler extends SimpleChannelUpstreamHandler {
    private var handler: RequestHandler = _

    override def messageReceived(context: ChannelHandlerContext, event: MessageEvent) = if (event.getMessage != null) {
      contextualize(context, event) {
        try {
          if (handler != null) {
            invokeHandler(context, event, handler)
          } else {
            lookupHandler(context, event, resources) match {
              case Some(h) => {
                handler = h
                invokeHandler(context, event, handler)
              }
              case None => notFound(context, event)
            }
          }
        } catch {
          case t: Throwable => NettyWebapp.this.exceptionCaught(context, t, handler)
        }
      }
    }

    override def exceptionCaught(ctx: ChannelHandlerContext, e: ExceptionEvent) {
      NettyWebapp.this.exceptionCaught(ctx, e.getCause, handler)
    }
  }
}

object NettyWebapp {
  private var instance: NettyWebapp = null

  def apply() = instance

  /**
   * Creates a URL from the given values. This method allows overriding the internally seen URL.
   */
  def createURL(method: Method,
                protocol: String,
                host: String,
                port: Int,
                path: String,
                parameters: Map[String, List[String]]) = {
    if (instance != null) {
      instance.createURL(method, protocol, host, port, path, parameters)
    } else {
      URL(method, protocol, host, port, path, parameters)
    }
  }
}