package com.outr.webcommunicator.netty.handler

import org.powerscala.Logging
import org.jboss.netty.channel._
import com.outr.webcommunicator.netty.NettyWebapp
import org.jboss.netty.handler.codec.http._
import java.util.{Date, TimeZone, Locale, Calendar}
import java.text.SimpleDateFormat
import java.io.{RandomAccessFile, File, InputStream}
import org.jboss.netty.handler.stream.{ChunkedFile, ChunkedStream}
import org.jboss.netty.handler.ssl.SslHandler
import java.net.URL
import javax.activation.MimetypesFileTypeMap
import org.jboss.netty.buffer.ChannelBuffers
import org.jboss.netty.util.CharsetUtil
import io.Source

/**
 * RequestHandler, as the name suggests, is responsible for handling requests.
 *
 * There are several streaming methods to help facilitate its use, but only the apply method need be implemented.
 *
 * @author Matt Hicks <mhicks@powerscala.org>
 */
trait RequestHandler extends Logging {
  /**
   * Determines the default chunkSize to use for streaming methods.
   *
   * Defaults to using RequestHandler.chunkSize
   */
  def chunkSize = RequestHandler.chunkSize

  /**
   * Determines the default cacheTimeInSeconds to use for streaming methods.
   *
   * Defaults to using RequestHandler.cacheTimeInSeconds
   */
  def cacheTimeInSeconds = RequestHandler.cacheTimeInSeconds

  /**
   * Determines the default mimeTypes to use for streaming methods when a content-type is not defined.
   *
   * Defaults to using RequestHandler.mimeTypes
   */
  def mimeTypes = RequestHandler.mimeTypes

  /**
   * Handles an incoming request.
   *
   * @param webapp the webapp caller
   * @param context the current context
   * @param event the MessageEvent in use
   */
  def apply(webapp: NettyWebapp, context: ChannelHandlerContext, event: MessageEvent): Unit

  def streamURL(url: URL,
                context: ChannelHandlerContext,
                request: HttpRequest,
                contentType: String = null,
                cacheTimeInSeconds: Int = cacheTimeInSeconds,
                chunkSize: Int = chunkSize,
                enableCaching: Boolean = true) = RequestHandler.streamURL(url, context, request, contentType, cacheTimeInSeconds, chunkSize, enableCaching)

  def streamInput(input: InputStream,
                  context: ChannelHandlerContext,
                  request: HttpRequest,
                  contentType: String,
                  contentLength: Long = -1,
                  lastModified: Long = -1,
                  cacheTimeInSeconds: Int = cacheTimeInSeconds,
                  chunkSize: Int = chunkSize,
                  enableCaching: Boolean = true) = RequestHandler.streamInput(input, context, request, contentType, contentLength, lastModified, cacheTimeInSeconds, chunkSize, enableCaching)

  def streamFile(file: File,
                 context: ChannelHandlerContext,
                 request: HttpRequest,
                 contentType: String = null,
                 cacheTimeInSeconds: Int = cacheTimeInSeconds,
                 chunkSize: Int = chunkSize,
                 enableCaching: Boolean = true) = RequestHandler.streamFile(file, context, request, contentType, cacheTimeInSeconds, chunkSize, enableCaching)

  def streamString(s: String,
                   context: ChannelHandlerContext,
                   request: HttpRequest,
                   contentType: String,
                   lastModified: Long = -1,
                   cacheTimeInSeconds: Int = cacheTimeInSeconds,
                   enableCaching: Boolean = true) = RequestHandler.streamString(s, context, request, contentType, lastModified, cacheTimeInSeconds, enableCaching)

  def errorThrown(t: Throwable) = error("Default handling of RequestHandler.errorThrown", t)
}

object RequestHandler {
  /**
   * The default chunk size to use for streaming methods in RequestHandler.
   *
   * Defaults to 8192
   */
  val chunkSize = 8192
  /**
   * The default cache time to use for streaming methods in RequestHandler.
   *
   * Defaults to 120
   */
  val cacheTimeInSeconds = 120
  /**
   * The default mimeTypes to use for streaming methods in RequestHandler.
   */
  val mimeTypes = new MimetypesFileTypeMap() {
    addMimeTypes(Source.fromURL(getClass.getClassLoader.getResource("mime_types.txt")).mkString)
  }

  def createDateFormatter: SimpleDateFormat = {     // DateFormatter isn't thread-safe
    val f = new SimpleDateFormat("EEE, dd MMM yyy HH:mm:ss zzz", Locale.US)
    f.setTimeZone(TimeZone.getTimeZone("GMT"))
    f
  }

  /**
   * Streams a URL back to the client and closes the connection when finished.
   *
   * @param url the URL to stream
   * @param context the context to use
   * @param request the inbound request
   * @param contentType the content-type to send. If this is null the content-type will be derived from the URL
   * @param cacheTimeInSeconds the amount of time in seconds to cache this response
   * @param chunkSize the chunk size to use when streaming the content back
   * @param enableCaching determines whether this content is cachable
   */
  def streamURL(url: URL,
                context: ChannelHandlerContext,
                request: HttpRequest,
                contentType: String = null,
                cacheTimeInSeconds: Int = cacheTimeInSeconds,
                chunkSize: Int = chunkSize,
                enableCaching: Boolean = true) = {
    val connection = url.openConnection()
    val lastModified = connection.getLastModified
    if (!enableCaching || !sendCached(lastModified, context, request)) {
      val contentLength = connection.getContentLength
      val input = connection.getInputStream
      val path = url.toString
      val filename = if (path.indexOf('/') != -1) {
        path.substring(path.lastIndexOf('/'))
      } else {
        path
      }
      val response = createResponse(contentLength = contentLength,
                                    contentType = getContentType(filename, contentType),
                                    cacheTimeInSeconds = cacheTimeInSeconds,
                                    lastModified = lastModified,
                                    sendExpiration = enableCaching)
      val channel = context.getChannel
      channel.write(response)
      writeInput(channel, input, chunkSize, closeOnFinish = true)
    }
  }

  /**
   * Streams an InputStream back to the client and closes the connection when finished.
   *
   * @param input the InputStream to stream
   * @param context the context to use
   * @param request the inbound request
   * @param contentType the content-type to send
   * @param contentLength the content-length to send
   * @param lastModified the last modified time in milliseconds
   * @param cacheTimeInSeconds the amount of time in seconds to cache this response
   * @param chunkSize the chunk size to use when streaming the content back
   * @param enableCaching determines whether this content is cachable
   */
  def streamInput(input: InputStream,
                  context: ChannelHandlerContext,
                  request: HttpRequest,
                  contentType: String,
                  contentLength: Long = -1,
                  lastModified: Long = -1,
                  cacheTimeInSeconds: Int = cacheTimeInSeconds,
                  chunkSize: Int = chunkSize,
                  enableCaching: Boolean = true) = {
    if (!enableCaching || !sendCached(lastModified, context, request)) {
      val response = createResponse(contentLength = contentLength,
                                    contentType = contentType,
                                    cacheTimeInSeconds = cacheTimeInSeconds,
                                    lastModified = lastModified,
                                    sendExpiration = enableCaching)
      val channel = context.getChannel
      channel.write(response)
      writeInput(channel, input, chunkSize, closeOnFinish = true)
    }
  }

  /**
   * Streams a File back to the client and closes the connection when finished.
   *
   * @param file the File to stream
   * @param context the context to use
   * @param request the inbound request
   * @param contentType the content-type to send. If this is null the content-type will be derived from the URL
   * @param cacheTimeInSeconds the amount of time in seconds to cache this response
   * @param chunkSize the chunk size to use when streaming the content back
   * @param enableCaching determines whether this content is cachable
   */
  def streamFile(file: File,
                 context: ChannelHandlerContext,
                 request: HttpRequest,
                 contentType: String = null,
                 cacheTimeInSeconds: Int = cacheTimeInSeconds,
                 chunkSize: Int = chunkSize,
                 enableCaching: Boolean = true) = {
    if (!enableCaching || !sendCached(file.lastModified(), context, request)) {
      val raf = new RandomAccessFile(file, "r")
      val response = createResponse(contentLength = raf.length(),
                                    contentType = getContentType(file.getName, contentType),
                                    cacheTimeInSeconds = cacheTimeInSeconds,
                                    lastModified = file.lastModified(),
                                    sendExpiration = enableCaching)
      val channel = context.getChannel
      channel.write(response)
      writeFile(channel, raf, chunkSize, closeOnFinish = true)
    }
  }

  /**
   * Streams a String back to the client and closes the connection when finished.
   *
   * @param s the String to stream
   * @param context the context to use
   * @param request the inbound request
   * @param contentType the content-type to send
   * @param lastModified the last modified time in milliseconds
   * @param cacheTimeInSeconds the amount of time in seconds to cache this response
   * @param enableCaching determines whether this content is cachable
   * @param status defines the HttpResponseStatus (default to OK)
   */
  def streamString(s: String,
                   context: ChannelHandlerContext,
                   request: HttpRequest,
                   contentType: String,
                   lastModified: Long = -1,
                   cacheTimeInSeconds: Int = cacheTimeInSeconds,
                   enableCaching: Boolean = true,
                   status: HttpResponseStatus = HttpResponseStatus.OK) = {
    if (!enableCaching || !sendCached(lastModified, context, request)) {
      val content = ChannelBuffers.copiedBuffer(s, CharsetUtil.UTF_8)
      val response = createResponse(status = status,
                                    contentLength = content.readableBytes(),
                                    contentType = "%s; charset=UTF-8".format(contentType),
                                    cacheTimeInSeconds = cacheTimeInSeconds,
                                    lastModified = lastModified,
                                    sendExpiration = enableCaching)
      response.setContent(content)
      val channel = context.getChannel
      channel.write(response).addListener(ChannelFutureListener.CLOSE)
    }
  }

  /**
   * Sends a standard OK response and closes the connection.
   */
  def sendResponse(context: ChannelHandlerContext) = {
    val response = createResponse()
    context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  /**
   * Attempts to use mimeTypes to determine the content-type if the supplied content-type is null.
   *
   * @param filename the filename to use for content-type derivation
   * @param contentType the content-type to use if not null
   * @return derived content-type
   */
  def getContentType(filename: String, contentType: String = null) = contentType match {
    case null => mimeTypes.getContentType(filename)
    case s => s
  }

  /**
   * Reads from the request the last modified date in the headers and determines if the version the client has the same
   * modification stamp as the server.
   *
   * Note that this will send a response back and close the connection if the client has the cached version.
   *
   * @param lastModified the last modified stamp in milliseconds to compare with
   * @param context the context to use
   * @param request the inbound request
   * @return true if the client has the correct version
   */
  def sendCached(lastModified: Long, context: ChannelHandlerContext, request: HttpRequest) = {
    val ifModifiedSince = request.getHeader(HttpHeaders.Names.IF_MODIFIED_SINCE)
    if (ifModifiedSince != null && ifModifiedSince != "" && lastModified != -1L) {
      val date = RequestHandler.createDateFormatter.parse(ifModifiedSince)
      val seconds = date.getTime / 1000
      val lastModifiedSeconds = lastModified / 1000
      if (seconds == lastModifiedSeconds) {
        val response = createResponse(status = HttpResponseStatus.NOT_MODIFIED, sendExpiration = false)
        context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
        true
      } else {
        false
      }
    } else {
      false
    }
  }

  /**
   * Creates an HttpResponse given the parameters supplied.
   *
   * @param version the HttpVersion, defaults to HTTP_1_1
   * @param status the HttpResponseStatus, defaults to OK
   * @param contentLength the content-length, defaults to -1 (not sent)
   * @param contentType the content-type, defaults to null (not sent)
   * @param cacheTimeInSeconds the time to cache the response in seconds
   * @param lastModified the last modified timestamp, defaults to -1 (not sent)
   * @param sendExpiration true if expiration data should be sent, defaults to true
   */
  def createResponse(version: HttpVersion = HttpVersion.HTTP_1_1,
                     status: HttpResponseStatus = HttpResponseStatus.OK,
                     contentLength: Long = -1,
                     contentType: String = null,
                     cacheTimeInSeconds: Int = cacheTimeInSeconds,
                     lastModified: Long = -1,
                     sendExpiration: Boolean = true) = {
    val r = new DefaultHttpResponse(version, status)
    if (contentLength != -1) {
      HttpHeaders.setContentLength(r, contentLength)
    }
    if (contentType != null) {
      r.setHeader(HttpHeaders.Names.CONTENT_TYPE, contentType)
    }
    val c = Calendar.getInstance()
    val f = RequestHandler.createDateFormatter
    r.setHeader(HttpHeaders.Names.DATE, f.format(c.getTime))
    if (sendExpiration) {
      c.add(Calendar.SECOND, cacheTimeInSeconds)
      r.setHeader(HttpHeaders.Names.EXPIRES, f.format(c.getTime))
      r.setHeader(HttpHeaders.Names.CACHE_CONTROL, "private, max-age=%s".format(cacheTimeInSeconds))
    }
    if (lastModified != -1L) {
      r.setHeader(HttpHeaders.Names.LAST_MODIFIED, f.format(new Date(lastModified)))
    }
    r
  }

  /**
   * Writes the InputStream out. This handles explicitly streaming data back to the channel. The streamInput method
   * should be used to send a response instead of this method.
   */
  def writeInput(channel: Channel, input: InputStream, chunkSize: Int = chunkSize, closeOnFinish: Boolean = true) = {
    if (input == null) {
      throw new NullPointerException("Input cannot be null")
    }
    val f = channel.write(new ChunkedStream(input, chunkSize))
    if (closeOnFinish) {
      f.addListener(new ChannelFutureListener {
        def operationComplete(future: ChannelFuture) = input.close()
      })
      f.addListener(ChannelFutureListener.CLOSE)
    }
    f
  }

  /**
   * Writes the RandomAccessFile out. This handles explicitly streaming data back to the channel. The streamFile method
   * should be used to send a response instead of this method.
   */
  def writeFile(channel: Channel, file: RandomAccessFile, chunkSize: Int = chunkSize, closeOnFinish: Boolean = true) = {
    val f = if (channel.getPipeline.get(classOf[SslHandler]) != null) {
      channel.write(new ChunkedFile(file, 0, file.length(), chunkSize))
    } else {
      val region = new DefaultFileRegion(file.getChannel, 0, file.length())
      val future = channel.write(region)
      future.addListener(new ChannelFutureListener {
        def operationComplete(future: ChannelFuture) {
          region.releaseExternalResources()
        }
      })
      future
    }
    if (closeOnFinish) {
      f.addListener(new ChannelFutureListener {
        def operationComplete(future: ChannelFuture) = file.close()
      })
      f.addListener(ChannelFutureListener.CLOSE)
    }
    f
  }

  def responder(status: HttpResponseStatus): RequestHandler = new RequestHandler {
    def apply(webapp: NettyWebapp, context: ChannelHandlerContext, event: MessageEvent) = {
      val response = RequestHandler.createResponse(status = status, contentType = "text/plain; charset=UTF-8")
      response.setContent(ChannelBuffers.copiedBuffer(response.getStatus.toString, CharsetUtil.UTF_8))
      context.getChannel.write(response).addListener(ChannelFutureListener.CLOSE)
    }
  }

  def redirect(url: String, context: ChannelHandlerContext) = {
    val s = "<html><head><meta http-equiv=\"refresh\" content=\"0;url=%s\"></head><body>redirect</body></html>".format(url)
    val content = ChannelBuffers.copiedBuffer(s, CharsetUtil.UTF_8)
    val response = createResponse(status = HttpResponseStatus.FOUND,
      contentLength = content.readableBytes(),
      contentType = "text/html; charset=UTF-8",
      cacheTimeInSeconds = cacheTimeInSeconds,
      sendExpiration = false)
    response.setHeader(HttpHeaders.Names.LOCATION, url)
    response.setContent(content)
    val channel = context.getChannel
    channel.write(response).addListener(ChannelFutureListener.CLOSE)
  }

  def redirector(url: String) = new RequestHandler {
    def apply(webapp: NettyWebapp, context: ChannelHandlerContext, event: MessageEvent) = {
      redirect(url, context)
    }
  }
}