package com.outr.webcommunicator.proxy

import org.jboss.netty.channel._
import org.jboss.netty.handler.codec.http.{HttpRequestEncoder, HttpRequest}
import org.jboss.netty.bootstrap.ClientBootstrap
import org.jboss.netty.buffer.ChannelBuffers
import org.powerscala.Logging

import com.outr.webcommunicator.netty._

/**
 * @author Matt Hicks <mhicks@outr.com>
 */
class ProxyHandler(server: ProxyServer) extends SimpleChannelUpstreamHandler with Logging {
  @volatile var inboundChannel: Channel = _
  @volatile var outboundChannel: Channel = _

  override def channelOpen(ctx: ChannelHandlerContext, e: ChannelStateEvent) {
    inboundChannel = e.getChannel
  }

  override def messageReceived(ctx: ChannelHandlerContext, evt: MessageEvent) = {
    val message = evt.getMessage
    message match {
      case request: HttpRequest if (outboundChannel == null) => {
        val address = server.findProxy(request)
        info("%s being proxied to %s:%s".format(request2URL(request), address.getHostName, address.getPort))
        val clientBootstrap = new ClientBootstrap(server.clientFactory)
        clientBootstrap.getPipeline.addLast("encoder", new HttpRequestEncoder)
        clientBootstrap.getPipeline.addLast("handler", new OutboundHandler(this))
        val future = clientBootstrap.connect(address)
        inboundChannel.setReadable(false)
        outboundChannel = future.getChannel

        future.addListener(new ChannelFutureListener {
          def operationComplete(future: ChannelFuture) = {
            if (future.isSuccess) {
              inboundChannel.setReadable(true)    // Connection successful, begin to accept incoming traffic

              outboundChannel.write(message).addListener(new ChannelFutureListener {
                def operationComplete(future: ChannelFuture) = {
                  // Remove the decoder and encoder after we finish writing back to support websockets
                  inboundChannel.getPipeline.remove("decoder")
                  outboundChannel.getPipeline.remove("encoder")
                }
              })
              // If outboundChannel is saturated, do not read until notified in OutboundHandler.channelInterestChanged
              if (!outboundChannel.isWritable) {
                evt.getChannel.setReadable(false)
              }
            } else {
              inboundChannel.close()              // Connection attempt failed, close the connection
            }
          }
        })
      }
      case _ => outboundChannel.write(message)
    }
  }

  override def channelInterestChanged(ctx: ChannelHandlerContext, e: ChannelStateEvent) {
    // If inboundChannel is not saturated anymore, continue accepting the incoming traffic from the outboundChannel
    synchronized {
      if (e.getChannel.isWritable) {
        if (outboundChannel != null) {
          outboundChannel.setReadable(true)
        }
      }
    }
  }

  override def channelClosed(ctx: ChannelHandlerContext, e: ChannelStateEvent) {
    if (outboundChannel != null) {
      ProxyHandler.closeOnFlush(outboundChannel)
    }
  }

  override def exceptionCaught(ctx: ChannelHandlerContext, e: ExceptionEvent) {
    e.getCause.printStackTrace()
    ProxyHandler.closeOnFlush(e.getChannel)
  }
}

object ProxyHandler {
  def closeOnFlush(channel: Channel) = {
    if (channel.isConnected) {
      channel.write(ChannelBuffers.EMPTY_BUFFER).addListener(ChannelFutureListener.CLOSE)
    }
  }
}