package com.outr.webcommunicator.proxy

import java.util.concurrent.Executors
import org.jboss.netty.bootstrap.ServerBootstrap
import org.jboss.netty.channel.socket.nio.{NioClientSocketChannelFactory, NioServerSocketChannelFactory}
import java.net.InetSocketAddress
import org.jboss.netty.channel.{Channels, ChannelPipelineFactory}
import org.jboss.netty.handler.codec.http.{HttpRequest, HttpRequestDecoder}

import com.outr.webcommunicator.netty._
import annotation.tailrec
import com.outr.webcommunicator.URL

/**
 * @author Matt Hicks <mhicks@outr.com>
 */
class ProxyServer(val host: String = null, val port: Int = 8080) extends ChannelPipelineFactory {
  private var _httpProxies = List.empty[HttpRequestProxy]

  def httpProxies = _httpProxies
  def +=(proxy: HttpRequestProxy) = synchronized {
    _httpProxies = (proxy :: _httpProxies.reverse).reverse
  }
  def -=(proxy: HttpRequestProxy) = synchronized {
    _httpProxies = _httpProxies.filterNot(p => p == proxy)
  }

  val address = host match {
    case null | "*" => new InetSocketAddress(port)
    case _ => new InetSocketAddress(host, port)
  }
  private val executor = Executors.newCachedThreadPool()
  private val bootstrap = new ServerBootstrap(new NioServerSocketChannelFactory(executor, executor))
  val clientFactory = new NioClientSocketChannelFactory(executor, executor)

  bootstrap.setPipelineFactory(this)
  bootstrap.bind(address)

  def getPipeline = {
    val pipeline = Channels.pipeline()
    pipeline.addLast("decoder", new HttpRequestDecoder)
    pipeline.addLast("handler", new ProxyHandler(this))
    pipeline
  }

  @tailrec
  final def findProxy(request: HttpRequest, proxies: List[HttpRequestProxy] = httpProxies): InetSocketAddress = {
    if (proxies.isEmpty) {
      throw new RuntimeException("Unable to find proxy for URL: %s".format(request2URL(request)))
    } else {
      val proxy = proxies.head
      proxy(request) match {
        case Some((remoteHost, remotePort)) => new InetSocketAddress(remoteHost, remotePort)
        case None => findProxy(request, proxies.tail)
      }
    }
  }
}

object ProxyServer {
  def main(args: Array[String]): Unit = {
    val server = new ProxyServer(port = 8080)
    server += new URLProxy {
      def apply(url: URL): Option[(String, Int)] = Some(("localhost", 8888))
    }
  }
}