// Avoid `console` errors in browsers that lack a console.
if (!(window.console && console.log)) {
    (function() {
        var noop = function() {};
        var methods = ['assert', 'clear', 'count', 'debug', 'dir', 'dirxml', 'error', 'exception', 'group', 'groupCollapsed', 'groupEnd', 'info', 'log', 'markTimeline', 'profile', 'profileEnd', 'markTimeline', 'table', 'time', 'timeEnd', 'timeStamp', 'trace', 'warn'];
        var length = methods.length;
        var console = window.console = {};
        while (length--) {
            console[methods[length]] = noop;
        }
    }());
}

jQuery(window).keypress(function(e) {
    if (e.keyCode == 27) {      // Avoid ESC killing the connection on Firefox
        e.preventDefault();
    }
});

// Creates a universally unique identifier used to communicate to the server
function createUUID() {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
        var r = Math.random()*16|0, v = c == 'x' ? r : (r&0x3|0x8);
        return v.toString(16);
    });
}

function Communicator() {
    var c = this;
    $(window).unload(function () {
        c.disconnect();
    });
}

Communicator.prototype = {
    /**
     * id is the unique identifier used to differentiate this client when talking to the server. This is generated upon
     * connect if not specified before.
     */
    id: null,
    idCreated: false,
    webSocketURL: null,
    pollingURL: null,
    sendingURL: null,
    timeout: 30000,
    retryDelay: 2000,
    maxRetries: 5,
    socket: null,
    ajaxPollingRequest: null,
    ajaxSendingRequest: null,
    failures: 0,
    sending: false,
    queue: [],
    sentQueue: [],
    lastReceiveId: 0,
    lastSendId: 0,
    connected: false,
    debug: false,
    bindings: {},
    /**
     * Binds a 'handler' function to messages received of type 'event'.
     *
     * @param handler
     */
    on: function(event, handler) {
        this.bindings[event] = handler;
    },
    /**
     * Invoked when a message is received from the server.
     *
     * @param event
     * @param message
     */
    onReceive: function(event, message) {
        var handler = this.bindings[event];
        if (handler != null) {
            handler(message);
        } else {
            log('Unhandled event received: ' + event + ' - ' + JSON.stringify(message));
        }
    },
    /**
     * Invoked when the connection has been established with a protocol.
     */
    onConnected: function() {
    },
    connect: function() {
        if (!window.WebSocket) {
            window.WebSocket = window.MozWebsocket;     // Support for older WebSocket support
        }
        if (this.webSocketURL != null && window.WebSocket) {
            log('Attempting to connect via WebSocket');
            this.connectWebSocket()
        } else if (this.pollingURL != null) {
            log('Attempting to connect via AJAX');
            this.connectPolling();
        } else {
            log('Neither webSocketURL or pollingURL are defined, so we cannot connect!');
        }
    },
    disconnect: function() {
        this.failures = this.maxRetries;        // Stop trying
        if (this.socket != null) {
            this.socket.close();
        }
        if (this.ajaxPollingRequest != null) {
            this.ajaxPollingRequest.abort();
        }
        if (this.ajaxSendingRequest != null) {
            this.ajaxSendingRequest.abort();
        }
    },
    createWebSocket: function() {
        this.socket = new WebSocket(this.webSocketURL);
        this.socket.onmessage = function(event) {
            c.onWebSocketMessage(event);
        };
        var c = this;
        this.socket.onclose = function(event) {
            c.onWebSocketClosed(event);
        };
    },
    connectWebSocket: function() {
        if (this.id == null) {
            this.id = createUUID();
            this.idCreated = true;
            if (this.debug) log('Connecting WebSocket (' + this.webSocketURL + ')...');
            this.createWebSocket();
            var c = this;
            this.socket.onopen = function(event) {
                c.send('create', c.id);        // Identify this WebSocket to an id
                c.connected = true;
                c.sendQueue();
                c.onWebSocketOpened(event);
                c.onConnected();
            };
            this.socket.onerror = function(error) {
                c.onWebSocketError(error);
            };
        } else {
            this.reconnectWebSocket();
        }
    },
    reconnectWebSocket: function() {
        if (this.debug) log('Reconnecting WebSocket (' + this.webSocketURL + ')...');
        this.createWebSocket();
        var c = this;
        this.socket.onopen = function(event) {
            var json = {
                uuid: c.id,
                lastReceiveId: c.lastReceiveId
            };
            c.send('reconnect', json);        // Identify this WebSocket to an id
            c.connected = true;
            c.sendQueue();
            c.onWebSocketOpened(event);
            c.onConnected();
        };
    },
    onWebSocketOpened: function(event) {
        log('WebSocket opened successfully!');
        this.failures = 0;      // Reset failures after a successful connect
    },
    onWebSocketError: function(error) {
    },
    onWebSocketMessage: function(event) {
        var json = jQuery.parseJSON(event.data);
        var id = json.id;
        if (id == this.lastReceiveId + 1) {
            this.onReceive(json.event, json.message);
            this.lastReceiveId = id;
        } else {
            log('Expected receive id of ' + (this.lastReceiveId + 1) + ' but received ' + id);
        }
    },
    onWebSocketClosed: function(event) {
        log('WebSocket closed!');
        this.socket = null;     // Remove the reference
        if (!this.connected && this.pollingURL != null) {
            // Switch to polling since we cannot connect to the websocket
            log('Failing over to AJAX polling since we are unable to connect via WebSocket');
            if (this.idCreated) {           // Reset the id
                this.idCreated = false;
                this.id = null;
            }
            this.connectPolling();
        } else {
            this.failures += 1;
            if (this.failures > this.maxRetries) {
                this.connected = false;
                log('Maximum retries reached, giving up!');
            } else {
                var c = this;
                setTimeout(function () {
                    c.reconnectWebSocket();
                }, this.retryDelay);
            }
        }
    },
    sendPolling: function(query) {
        var c = this;
        if (this.debug) log('sendPolling: ' + this.pollingURL + ' - ' + JSON.stringify(query));
        this.ajaxPollingRequest = $.ajax({
            type: 'POST',
            url: this.pollingURL,
            dataType: 'json',
            data: JSON.stringify(query),
            processData: false,
            cache: false,
            success: function(data) {
                c.onPollMessage(data);
            },
            error: function(jqXHR, textStatus, errorThrown) {
                c.onPollError(jqXHR, textStatus, errorThrown);
            },
            complete: function(event) {
                c.onPollComplete(event);
            },
            timeout: c.timeout
        });
        this.sendQueue();
    },
    connectPolling: function() {
        if (this.id == null) {
            if (this.debug) log('No existing id, creating a new one');
            this.id = createUUID();
            var query = {
                event: 'create',
                message: this.id
            };
            this.connected = true;
            this.sendPolling(query);
            this.onConnected();
        } else {
            if (this.debug) log('Existing id found, reconnecting: ' + this.id);
            this.reconnectPolling();
        }
    },
    reconnectPolling: function() {
        var query = {
            event: 'reconnect',
            message: JSON.stringify({
                uuid: this.id,
                lastReceiveId: this.lastReceiveId
            })
        };
        var firstConnection = !this.connected;
        this.connected = true;
        this.sendPolling(query);
        if (firstConnection) {
            this.onConnected();
        }
    },
    onPollMessage: function(data) {
        if (data.event == 'noop') {
            // Valid response, but nothing to receive
        } else {
            for (var i = 0; i < data.length; i++) {
                var event = data[i];
                var id = event.id;
                if (id == this.lastReceiveId + 1) {
                    this.onReceive(event.event, event.message);
                    this.lastReceiveId = id;
                } else {
                    log('Expected receive id of ' + (this.lastReceiveId + 1) + ' but received ' + id);
                }
            }
        }
        this.reconnectPolling();
    },
    onPollError: function(jqXHR, textStatus, errorThrown) {
        log('Polling error received, will retry after delay (' + textStatus + ' / ' + errorThrown + ')...');
        this.failures += 1;
        if (this.failures > this.maxRetries) {
            this.connected = false;
            log('Maximum retries reached (' + this.failures + '), giving up!');
        } else {
            var c = this;
            setTimeout(function () {
                c.reconnectPolling();
            }, this.retryDelay);
        }
    },
    onPollComplete: function(event) {
        this.ajaxPollingRequest = null;
    },
    send: function(event, message) {
        var t = typeof message;
        if (t != 'string') {
            message = JSON.stringify(message);
        }
        var messageId = ++this.lastSendId;
        var json = {
            id: messageId,
            event: event,
            message: message
        };
        if (this.socket != null && this.socket.readyState == WebSocket.OPEN) {
            this.sendWebSocket(json);
        } else {
            this.queue.push(json);
            this.sendQueue();
        }
    },
    sendWebSocket: function(message) {
        this.socket.send(JSON.stringify(message));
    },
    sendQueue: function() {
        if (!this.sending && this.queue.length > 0 && this.connected) {
            this.sending = true;
            if (this.socket != null) {
                try {
                    if (this.socket.readyState == WebSocket.OPEN) {
                        this.sentQueue = this.queue;
                        this.queue = [];
                        for (var i = 0; i < this.sentQueue.length; i++) {
                            this.sendWebSocket(this.sentQueue[i]);
                        }
                    }
                } finally {
                    this.sending = false;
                }
            } else if (this.sendingURL != null) {
                this.sentQueue = this.queue;
                this.queue = [];
                var data = {
                    uuid: this.id,
                    messages: this.sentQueue
                };
                var c = this;
                if (this.debug) log('Sending queue: ' + JSON.stringify(data));
                this.ajaxSendingRequest = $.ajax({
                    type: 'POST',
                    url: this.sendingURL,
                    dataType: 'json',
                    data: JSON.stringify(data),
                    processData: false,
                    cache: false,
                    success: function(data) {
                        c.onAJAXSendSuccess(data);
                    },
                    error: function(jqXHR, textStatus, errorThrown) {
                        c.onAJAXSendError(jqXHR, textStatus, errorThrown);
                    },
                    complete: function() {
                        c.onAJAXSendComplete();
                    },
                    timeout: c.timeout
                });
            } else {        // Nothing connected yet
                this.sending = false;
            }
        }
    },
    onAJAXSendSuccess: function(data) {
        this.sending = false;
        if (this.queue.length > 0) {
            this.sendQueue();
        }
    },
    onAJAXSendError: function(jqXHR, textStatus, errorThrown) {
        this.sending = false;
        this.queue = this.sentQueue.concat(this.queue);        // Put the unsent data back into the queue
        log('Error while sending to the server via AJAX: ' + textStatus + ', Error: ' + errorThrown);
        // TODO: delay before resending
    },
    onAJAXSendComplete: function() {
        this.ajaxSendingRequest = null;
    }
};

function log(msg) {
    var message = new Date().toLocaleString() + ': ' + msg;
    console.log(message);
}